@@ -1,5 +1,5 @@
 /*
- * known_hosts.c
+ * pki.c
  * This file is part of the SSH Library
  *
  * Copyright (c) 2010 by Aris Adamantiadis
@@ -98,6 +98,25 @@ enum ssh_keytypes_e pki_privatekey_type_from_string(const char *privkey) {
     return SSH_KEYTYPE_UNKNOWN;
 }
 
+/**
+ * @brief returns the ECDSA key name ("ecdsa-sha2-nistp256" for example)
+ *
+ * @param[in] key the ssh_key whose ECDSA name to get
+ *
+ * @returns the ECDSA key name ("ecdsa-sha2-nistp256" for example)
+ *
+ * @returns "unknown" if the ECDSA key name is not known
+ */
+const char *ssh_pki_key_ecdsa_name(const ssh_key key)
+{
+#ifdef HAVE_OPENSSL_ECC /* FIXME Better ECC check needed */
+    return pki_key_ecdsa_nid_to_name(key->ecdsa_nid);
+#else
+    (void) key; /* unused */
+    return NULL;
+#endif
+}
+
 /**
  * @brief creates a new empty SSH key
  * @returns an empty ssh_key handle, or NULL on error.
@@ -138,6 +157,11 @@ void ssh_key_clean (ssh_key key){
     if(key->ecdsa) EC_KEY_free(key->ecdsa);
 #endif /* HAVE_OPENSSL_ECC */
 #endif
+    if (key->ed25519_privkey != NULL){
+        BURN_BUFFER(key->ed25519_privkey, sizeof(ed25519_privkey));
+        SAFE_FREE(key->ed25519_privkey);
+    }
+    SAFE_FREE(key->ed25519_pubkey);
     key->flags=SSH_KEY_FLAG_EMPTY;
     key->type=SSH_KEYTYPE_UNKNOWN;
     key->ecdsa_nid = 0;
@@ -188,6 +212,8 @@ const char *ssh_key_type_to_char(enum ssh_keytypes_e type) {
       return "ssh-rsa1";
     case SSH_KEYTYPE_ECDSA:
       return "ssh-ecdsa";
+    case SSH_KEYTYPE_ED25519:
+      return "ssh-ed25519";
     case SSH_KEYTYPE_UNKNOWN:
       return NULL;
   }
@@ -226,6 +252,8 @@ enum ssh_keytypes_e ssh_key_type_from_name(const char *name) {
             || strcmp(name, "ecdsa-sha2-nistp384") == 0
             || strcmp(name, "ecdsa-sha2-nistp521") == 0) {
         return SSH_KEYTYPE_ECDSA;
+    } else if (strcmp(name, "ssh-ed25519") == 0){
+        return SSH_KEYTYPE_ED25519;
     }
 
     return SSH_KEYTYPE_UNKNOWN;
@@ -243,7 +271,7 @@ int ssh_key_is_public(const ssh_key k) {
         return 0;
     }
 
-    return (k->flags & SSH_KEY_FLAG_PUBLIC);
+    return (k->flags & SSH_KEY_FLAG_PUBLIC) == SSH_KEY_FLAG_PUBLIC;
 }
 
 /**
@@ -258,7 +286,7 @@ int ssh_key_is_private(const ssh_key k) {
         return 0;
     }
 
-    return (k->flags & SSH_KEY_FLAG_PRIVATE);
+    return (k->flags & SSH_KEY_FLAG_PRIVATE) == SSH_KEY_FLAG_PRIVATE;
 }
 
 /**
@@ -281,7 +309,7 @@ int ssh_key_cmp(const ssh_key k1,
     }
 
     if (k1->type != k2->type) {
-        ssh_pki_log("key types don't macth!");
+        ssh_pki_log("key types don't match!");
         return 1;
     }
 
@@ -292,6 +320,10 @@ int ssh_key_cmp(const ssh_key k1,
         }
     }
 
+    if (k1->type == SSH_KEYTYPE_ED25519) {
+        return pki_ed25519_key_cmp(k1, k2, what);
+    }
+
     return pki_key_compare(k1, k2, what);
 }
 
@@ -335,6 +367,9 @@ void ssh_signature_free(ssh_signature sig)
             ECDSA_SIG_free(sig->ecdsa_sig);
 #endif
             break;
+        case SSH_KEYTYPE_ED25519:
+            SAFE_FREE(sig->ed25519_sig);
+            break;
         case SSH_KEYTYPE_UNKNOWN:
             break;
     }
@@ -367,6 +402,7 @@ int ssh_pki_import_privkey_base64(const char *b64_key,
                                   ssh_key *pkey)
 {
     ssh_key key;
+    int cmp;
 
     if (b64_key == NULL || pkey == NULL) {
         return SSH_ERROR;
@@ -379,7 +415,20 @@ int ssh_pki_import_privkey_base64(const char *b64_key,
     ssh_pki_log("Trying to decode privkey passphrase=%s",
                 passphrase ? "true" : "false");
 
-    key = pki_private_key_from_base64(b64_key, passphrase, auth_fn, auth_data);
+    /* Test for OpenSSH key format first */
+    cmp = strncmp(b64_key, OPENSSH_HEADER_BEGIN, strlen(OPENSSH_HEADER_BEGIN));
+    if (cmp == 0) {
+        key = ssh_pki_openssh_privkey_import(b64_key,
+                                             passphrase,
+                                             auth_fn,
+                                             auth_data);
+    } else {
+        /* fallback on PEM decoder */
+        key = pki_private_key_from_base64(b64_key,
+                                          passphrase,
+                                          auth_fn,
+                                          auth_data);
+    }
     if (key == NULL) {
         return SSH_ERROR;
     }
@@ -416,7 +465,6 @@ int ssh_pki_import_privkey_file(const char *filename,
                                 ssh_key *pkey) {
     struct stat sb;
     char *key_buf;
-    ssh_key key;
     FILE *file;
     off_t size;
     int rc;
@@ -470,18 +518,19 @@ int ssh_pki_import_privkey_file(const char *filename,
     }
     key_buf[size] = 0;
 
-    key = pki_private_key_from_base64(key_buf, passphrase, auth_fn, auth_data);
-    SAFE_FREE(key_buf);
-    if (key == NULL) {
-        return SSH_ERROR;
-    }
+    rc = ssh_pki_import_privkey_base64(key_buf,
+                                       passphrase,
+                                       auth_fn,
+                                       auth_data,
+                                       pkey);
 
-    *pkey = key;
-    return SSH_OK;
+    SAFE_FREE(key_buf);
+    return rc;
 }
 
 /**
- * @brief Export a private key to a pam file on disk.
+ * @brief Export a private key to a pem file on disk, or OpenSSH format for
+ *        keytype ssh-ed25519
  *
  * @param[in]  privkey  The private key to export.
  *
@@ -517,11 +566,17 @@ int ssh_pki_export_privkey_file(const ssh_key privkey,
         return SSH_EOF;
     }
 
-
-    blob = pki_private_key_to_pem(privkey,
-                                  passphrase,
-                                  auth_fn,
-                                  auth_data);
+    if (privkey->type == SSH_KEYTYPE_ED25519){
+        blob = ssh_pki_openssh_privkey_export(privkey,
+                                              passphrase,
+                                              auth_fn,
+                                              auth_data);
+    } else {
+        blob = pki_private_key_to_pem(privkey,
+                                      passphrase,
+                                      auth_fn,
+                                      auth_data);
+    }
     if (blob == NULL) {
         fclose(fp);
         return -1;
@@ -724,10 +779,36 @@ static int pki_import_pubkey_buffer(ssh_buffer buffer,
                 if (rc < 0) {
                     goto fail;
                 }
+
+                /* Update key type */
+                key->type_c = ssh_pki_key_ecdsa_name(key);
             }
             break;
 #endif
+        case SSH_KEYTYPE_ED25519:
+        {
+            ssh_string pubkey = buffer_get_ssh_string(buffer);
+            if (ssh_string_len(pubkey) != ED25519_PK_LEN) {
+                ssh_pki_log("Invalid public key length");
+                ssh_string_burn(pubkey);
+                ssh_string_free(pubkey);
+                goto fail;
+            }
+
+            key->ed25519_pubkey = malloc(ED25519_PK_LEN);
+            if (key->ed25519_pubkey == NULL) {
+                ssh_string_burn(pubkey);
+                ssh_string_free(pubkey);
+                goto fail;
+            }
+
+            memcpy(key->ed25519_pubkey, ssh_string_data(pubkey), ED25519_PK_LEN);
+            ssh_string_burn(pubkey);
+            ssh_string_free(pubkey);
+        }
+        break;
         case SSH_KEYTYPE_UNKNOWN:
+        default:
             ssh_pki_log("Unknown public key protocol %d", type);
             goto fail;
     }
@@ -815,7 +896,7 @@ int ssh_pki_import_pubkey_blob(const ssh_string key_blob,
         return SSH_ERROR;
     }
 
-    rc = buffer_add_data(buffer, ssh_string_data(key_blob),
+    rc = ssh_buffer_add_data(buffer, ssh_string_data(key_blob),
             ssh_string_len(key_blob));
     if (rc < 0) {
         ssh_pki_log("Out of memory!");
@@ -980,10 +1061,20 @@ int ssh_pki_generate(enum ssh_keytypes_e type, int parameter,
         case SSH_KEYTYPE_ECDSA:
 #ifdef HAVE_ECC
             rc = pki_key_generate_ecdsa(key, parameter);
-            if(rc == SSH_ERROR)
+            if (rc == SSH_ERROR) {
                 goto error;
+            }
+
+            /* Update key type */
+            key->type_c = ssh_pki_key_ecdsa_name(key);
             break;
 #endif
+        case SSH_KEYTYPE_ED25519:
+            rc = pki_key_generate_ed25519(key);
+            if (rc == SSH_ERROR) {
+                goto error;
+            }
+            break;
         case SSH_KEYTYPE_UNKNOWN:
             goto error;
     }
@@ -1062,11 +1153,11 @@ int ssh_pki_export_pubkey_blob(const ssh_key key,
 }
 
 /**
- * @brief Convert a public key to a base64 hased key.
+ * @brief Convert a public key to a base64 encoded key.
  *
  * @param[in] key       The key to hash
  *
- * @param[out] b64_key  A pointer to store the allocated base64 hashed key. You
+ * @param[out] b64_key  A pointer to store the allocated base64 encoded key. You
  *                      need to free the buffer.
  *
  * @return              SSH_OK on success, SSH_ERROR on error.
@@ -1240,9 +1331,9 @@ int ssh_pki_import_signature_blob(const ssh_string sig_blob,
         return SSH_ERROR;
     }
 
-    rc = buffer_add_data(buf,
-                         ssh_string_data(sig_blob),
-                         ssh_string_len(sig_blob));
+    rc = ssh_buffer_add_data(buf,
+                             ssh_string_data(sig_blob),
+                             ssh_string_len(sig_blob));
     if (rc < 0) {
         ssh_buffer_free(buf);
         return SSH_ERROR;
@@ -1299,12 +1390,19 @@ int ssh_pki_signature_verify_blob(ssh_session session,
 
         evp(key->ecdsa_nid, digest, dlen, ehash, &elen);
 
+#ifdef DEBUG_CRYPTO
+        ssh_print_hexa("Hash to be verified with ecdsa",
+                       ehash, elen);
+#endif
+
         rc = pki_signature_verify(session,
                                   sig,
                                   key,
                                   ehash,
                                   elen);
 #endif
+    } else if (key->type == SSH_KEYTYPE_ED25519) {
+        rc = pki_signature_verify(session, sig, key, digest, dlen);
     } else {
         unsigned char hash[SHA_DIGEST_LEN] = {0};
 
@@ -1334,7 +1432,7 @@ ssh_string ssh_pki_do_sign(ssh_session session,
     struct ssh_crypto_struct *crypto =
         session->current_crypto ? session->current_crypto :
                                   session->next_crypto;
-    ssh_signature sig;
+    ssh_signature sig = NULL;
     ssh_string sig_blob;
     ssh_string session_id;
     int rc;
@@ -1365,8 +1463,36 @@ ssh_string ssh_pki_do_sign(ssh_session session,
         evp_update(ctx, buffer_get_rest(sigbuf), buffer_get_rest_len(sigbuf));
         evp_final(ctx, ehash, &elen);
 
+#ifdef DEBUG_CRYPTO
+        ssh_print_hexa("Hash being signed", ehash, elen);
+#endif
+
         sig = pki_do_sign(privkey, ehash, elen);
 #endif
+    } else if (privkey->type == SSH_KEYTYPE_ED25519){
+        ssh_buffer buf;
+
+        buf = ssh_buffer_new();
+        if (buf == NULL) {
+            ssh_string_free(session_id);
+            return NULL;
+        }
+
+        ssh_buffer_set_secure(buf);
+        rc = ssh_buffer_pack(buf,
+                             "SP",
+                             session_id,
+                             buffer_get_rest_len(sigbuf), buffer_get_rest(sigbuf));
+        if (rc != SSH_OK) {
+            ssh_string_free(session_id);
+            ssh_buffer_free(buf);
+            return NULL;
+        }
+
+        sig = pki_do_sign(privkey,
+                          ssh_buffer_get_begin(buf),
+                          ssh_buffer_get_len(buf));
+        ssh_buffer_free(buf);
     } else {
         unsigned char hash[SHA_DIGEST_LEN] = {0};
         SHACTX ctx;
@@ -1458,10 +1584,8 @@ ssh_string ssh_srv_pki_do_sign_sessionid(ssh_session session,
                                          const ssh_key privkey)
 {
     struct ssh_crypto_struct *crypto;
-    unsigned char hash[SHA_DIGEST_LEN] = {0};
     ssh_signature sig;
     ssh_string sig_blob;
-    SHACTX ctx;
     int rc;
 
     if (session == NULL || privkey == NULL || !ssh_key_is_private(privkey)) {
@@ -1470,24 +1594,64 @@ ssh_string ssh_srv_pki_do_sign_sessionid(ssh_session session,
     crypto = session->next_crypto ? session->next_crypto :
                                        session->current_crypto;
 
-    ctx = sha1_init();
-    if (ctx == NULL) {
-        return NULL;
-    }
     if (crypto->secret_hash == NULL){
         ssh_set_error(session,SSH_FATAL,"Missing secret_hash");
         return NULL;
     }
-    sha1_update(ctx, crypto->secret_hash, crypto->digest_len);
-    sha1_final(hash, ctx);
+
+    if (privkey->type == SSH_KEYTYPE_ECDSA) {
+#ifdef HAVE_ECC
+        unsigned char ehash[EVP_DIGEST_LEN] = {0};
+        uint32_t elen;
+
+        evp(privkey->ecdsa_nid, crypto->secret_hash, crypto->digest_len,
+            ehash, &elen);
 
 #ifdef DEBUG_CRYPTO
-    ssh_print_hexa("Hash being signed", hash, SHA_DIGEST_LEN);
+        ssh_print_hexa("Hash being signed", ehash, elen);
 #endif
 
-    sig = pki_do_sign_sessionid(privkey, hash, SHA_DIGEST_LEN);
-    if (sig == NULL) {
-        return NULL;
+        sig = pki_do_sign_sessionid(privkey, ehash, elen);
+        if (sig == NULL) {
+            return NULL;
+        }
+#endif
+    } else if (privkey->type == SSH_KEYTYPE_ED25519) {
+        sig = ssh_signature_new();
+        if (sig == NULL){
+            return NULL;
+        }
+
+        sig->type = privkey->type;
+        sig->type_c = privkey->type_c;
+
+        rc = pki_ed25519_sign(privkey,
+                              sig,
+                              crypto->secret_hash,
+                              crypto->digest_len);
+        if (rc != SSH_OK){
+            ssh_signature_free(sig);
+            sig = NULL;
+        }
+    } else {
+        unsigned char hash[SHA_DIGEST_LEN] = {0};
+        SHACTX ctx;
+
+        ctx = sha1_init();
+        if (ctx == NULL) {
+            return NULL;
+        }
+        sha1_update(ctx, crypto->secret_hash, crypto->digest_len);
+        sha1_final(hash, ctx);
+
+#ifdef DEBUG_CRYPTO
+        ssh_print_hexa("Hash being signed", hash, SHA_DIGEST_LEN);
+#endif
+
+        sig = pki_do_sign_sessionid(privkey, hash, SHA_DIGEST_LEN);
+        if (sig == NULL) {
+            return NULL;
+        }
     }
 
     rc = ssh_pki_export_signature_blob(sig, &sig_blob);
